@@ -3,7 +3,8 @@
 #define _GDFONTS_H_ 1
 
 #ifdef __cplusplus
-extern "C" {
+extern "C"
+{
 #endif
 
 /*
@@ -18,11 +19,11 @@ extern "C" {
 
 #include "gd.h"
 
-extern gdFontPtr gdFontSmall;
+   BGD_EXPORT gdFontPtr gdFontSmall;
+   BGD_EXPORT gdFontPtr gdFontGetSmall(void);
 
 #ifdef __cplusplus
 }
 #endif
 
 #endif
-
