@@ -2,9 +2,13 @@
 /********************************************/
 /* gd interface to freetype library         */
 /*                                          */
-/* John Ellson   ellson@lucent.com          */
+/* John Ellson   ellson@graphviz.org        */
 /********************************************/
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -12,10 +16,12 @@
 #include "gd.h"
 #include "gdhelpers.h"
 
-#ifndef MSWIN32
+/* 2.0.10: WIN32, not MSWIN32 */
+#ifndef WIN32
 #include <unistd.h>
 #else
-#define R_OK 2
+#include <io.h>
+#define R_OK 04			/* Needed in Windows */
 #endif
 
 /* number of antialised colors for indexed bitmaps */
@@ -25,11 +31,20 @@ char *
 gdImageStringTTF (gdImage * im, int *brect, int fg, char *fontlist,
 		  double ptsize, double angle, int x, int y, char *string)
 {
-  gdImageStringFT (im, brect, fg, fontlist, ptsize,
-		   angle, x, y, string);
+  /* 2.0.6: valid return */
+  return gdImageStringFT (im, brect, fg, fontlist, ptsize,
+			  angle, x, y, string);
 }
 
 #ifndef HAVE_LIBFREETYPE
+char *
+gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
+		   double ptsize, double angle, int x, int y, char *string,
+		   gdFTStringExtraPtr strex)
+{
+  return "libgd was not built with FreeType font support\n";
+}
+
 char *
 gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 		 double ptsize, double angle, int x, int y, char *string)
@@ -39,8 +54,11 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 #else
 
 #include "gdcache.h"
-#include "freetype/freetype.h"
-#include "freetype/ftglyph.h"
+/* 2.0.16 Christophe Thomas: starting with FreeType 2.1.6, this is
+  mandatory, and it has been supported for a long while. */
+#include <ft2build.h>
+#include FT_FREETYPE_H
+#include FT_GLYPH_H
 
 /* number of fonts cached before least recently used is replaced */
 #define FONTCACHESIZE 6
@@ -57,19 +75,24 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 
 /*
  * The character (space) used to separate alternate fonts in the
- * fontlist parameter to gdImageStringFT.
+ * fontlist parameter to gdImageStringFT. 2.0.18: space was a
+ * poor choice for this.
  */
-#define LISTSEPARATOR " "
+#define LISTSEPARATOR ";"
 
 /*
  * DEFAULT_FONTPATH and PATHSEPARATOR are host type dependent and
- * are normally set by configure in gvconfig.h.  These are just
+ * are normally set by configure in config.h.  These are just
  * some last resort values that might match some Un*x system
  * if building this version of gd separate from graphviz.
  */
 #ifndef DEFAULT_FONTPATH
+#if defined(__APPLE__) || (defined(__MWERKS__) && defined(macintosh))
+#define DEFAULT_FONTPATH "/usr/share/fonts/truetype:/System/Library/Fonts:/Library/Fonts"
+#else
 #define DEFAULT_FONTPATH "/usr/share/fonts/truetype"
 #endif
+#endif
 #ifndef PATHSEPARATOR
 #define PATHSEPARATOR ":"
 #endif
@@ -94,29 +117,29 @@ typedef struct
 font_t;
 
 typedef struct
-  {
-    char *fontlist;		/* key */
-    FT_Library *library;
-  }
+{
+  char *fontlist;		/* key */
+  FT_Library *library;
+}
 fontkey_t;
 
 typedef struct
-  {
-    int pixel;			/* key */
-    int bgcolor;		/* key */
-    int fgcolor;		/* key *//* -ve means no antialias */
-    gdImagePtr im;		/* key */
-    int tweencolor;
-  }
+{
+  int pixel;			/* key */
+  int bgcolor;			/* key */
+  int fgcolor;			/* key *//* -ve means no antialias */
+  gdImagePtr im;		/* key */
+  int tweencolor;
+}
 tweencolor_t;
 
 typedef struct
-  {
-    int pixel;			/* key */
-    int bgcolor;		/* key */
-    int fgcolor;		/* key *//* -ve means no antialias */
-    gdImagePtr im;		/* key */
-  }
+{
+  int pixel;			/* key */
+  int bgcolor;			/* key */
+  int fgcolor;			/* key *//* -ve means no antialias */
+  gdImagePtr im;		/* key */
+}
 tweencolorkey_t;
 
 /********************************************************************
@@ -245,8 +268,7 @@ gdTcl_UtfToUniChar (char *str, Tcl_UniChar * chPtr)
 	   * by a trail-byte.
 	   */
 
-	  *chPtr = (Tcl_UniChar) (((byte & 0x1F) << 6)
-				  | (str[1] & 0x3F));
+	  *chPtr = (Tcl_UniChar) (((byte & 0x1F) << 6) | (str[1] & 0x3F));
 	  return 2;
 	}
       /*
@@ -267,7 +289,7 @@ gdTcl_UtfToUniChar (char *str, Tcl_UniChar * chPtr)
 	   */
 
 	  *chPtr = (Tcl_UniChar) (((byte & 0x0F) << 12)
-				| ((str[1] & 0x3F) << 6) | (str[2] & 0x3F));
+				  | ((str[1] & 0x3F) << 6) | (str[2] & 0x3F));
 	  return 3;
 	}
       /*
@@ -331,7 +353,7 @@ fontFetch (char **error, void *key)
   int n;
   int font_found = 0;
   unsigned short platform, encoding;
-  char *fontsearchpath, *fontpath, *fontlist;
+  char *fontsearchpath, *fontlist;
   char *fullname = NULL;
   char *name, *path, *dir;
   char *strtok_ptr;
@@ -349,7 +371,6 @@ fontFetch (char **error, void *key)
   fontsearchpath = getenv ("GDFONTPATH");
   if (!fontsearchpath)
     fontsearchpath = DEFAULT_FONTPATH;
-  path = strdup (fontsearchpath);
   fontlist = strdup (a->fontlist);
 
   /*
@@ -359,19 +380,26 @@ fontFetch (char **error, void *key)
        name = gd_strtok_r (0, LISTSEPARATOR, &strtok_ptr))
     {
 
+      /* make a fresh copy each time - strtok corrupts it. */
+      path = strdup (fontsearchpath);
       /*
        * Allocate an oversized buffer that is guaranteed to be
        * big enough for all paths to be tested.
        */
+      /* 2.0.22: Thorben Kundinger: +8 is needed, not +6. */
       fullname = gdRealloc (fullname,
-			    strlen (fontsearchpath) + strlen (name) + 6);
-      /* if name is an absolute filename then test directly */
-      if (*name == '/')
+			    strlen (fontsearchpath) + strlen (name) + 8);
+      /* if name is an absolute or relative pathname then test directly */
+      if (strchr (name, '/')
+	  || (name[0] != 0 && name[1] == ':'
+	      && (name[2] == '/' || name[2] == '\\')))
 	{
 	  sprintf (fullname, "%s", name);
 	  if (access (fullname, R_OK) == 0)
 	    {
 	      font_found++;
+	      /* 2.0.16: memory leak fixed, Gustavo Scotti */
+	      gdFree (path);
 	      break;
 	    }
 	}
@@ -384,27 +412,63 @@ fontFetch (char **error, void *key)
 	      font_found++;
 	      break;
 	    }
+	  sprintf (fullname, "%s/%s.pfa", dir, name);
+	  if (access (fullname, R_OK) == 0)
+	    {
+	      font_found++;
+	      break;
+	    }
+	  sprintf (fullname, "%s/%s.pfb", dir, name);
+	  if (access (fullname, R_OK) == 0)
+	    {
+	      font_found++;
+	      break;
+	    }
+	  sprintf (fullname, "%s/%s.dfont", dir, name);
+	  if (access (fullname, R_OK) == 0)
+	    {
+	      font_found++;
+	      break;
+	    }
 	}
+      gdFree (path);
       if (font_found)
 	break;
     }
-  gdFree (path);
   gdFree (fontlist);
   if (!font_found)
     {
       *error = "Could not find/open font";
+      /* 2.0.12: TBB: free these. Thanks to Frank Faubert. */
+      free (a->fontlist);
+      free (fullname);
+      gdFree (a);
+
       return NULL;
     }
 
   err = FT_New_Face (*b->library, fullname, 0, &a->face);
   if (err)
     {
+      /* 2.0.12: TBB: free these. Thanks to Frank Faubert. */
+      free (a->fontlist);
+      free (fullname);
+      gdFree (a);
       *error = "Could not read font";
       return NULL;
     }
   gdFree (fullname);
 
 /* FIXME - This mapping stuff is imcomplete - where is the spec? */
+/* EAM   - It's worse than that. It's pointless to match character encodings here. */
+/*         As currently written, the stored a->face->charmap only matches one of   */
+/*         the actual charmaps and we cannot know at this stage if it is the right */
+/*         one. We should just skip all this stuff, and check in gdImageStringFTEx */
+/*         if some particular charmap is preferred and if so whether it is held in */
+/*         one of the a->face->charmaps[0..num_charmaps].                          */
+/*         And why is it so bad not to find any recognized charmap?  The user may  */
+/*         still know what mapping to use, even if we do not.  In that case we can */
+/*         just use the map in a->face->charmaps[num_charmaps] and be done with it.*/
 
   a->have_char_map_unicode = 0;
   a->have_char_map_big5 = 0;
@@ -415,6 +479,24 @@ fontFetch (char **error, void *key)
       charmap = a->face->charmaps[n];
       platform = charmap->platform_id;
       encoding = charmap->encoding_id;
+
+/* EAM DEBUG - Newer versions of libfree2 make it easier 
+	by defining encodings */
+/* TBB: get this exactly right: 2.1.3 *or better*, all possible cases. */
+
+#if ((defined(FREETYPE_MAJOR)) && (((FREETYPE_MAJOR == 2) && (((FREETYPE_MINOR == 1) && (FREETYPE_PATCH >= 3)) || (FREETYPE_MINOR > 1))) || (FREETYPE_MAJOR > 2)))
+      if (charmap->encoding == FT_ENCODING_MS_SYMBOL
+	  || charmap->encoding == FT_ENCODING_ADOBE_CUSTOM
+	  || charmap->encoding == FT_ENCODING_ADOBE_STANDARD)
+	{
+	  a->have_char_map_unicode = 1;
+	  found = charmap;
+	  a->face->charmap = charmap;
+	  return (void *) a;
+	}
+#endif /* Freetype 2.1 or better */
+/* EAM DEBUG */
+
       if ((platform == 3 && encoding == 1)	/* Windows Unicode */
 	  || (platform == 3 && encoding == 0)	/* Windows Symbol */
 	  || (platform == 2 && encoding == 1)	/* ISO Unicode */
@@ -433,7 +515,7 @@ fontFetch (char **error, void *key)
 	  a->have_char_map_sjis = 1;
 	  found = charmap;
 	}
-      else if ((platform == 1 && encoding == 0)		/* Apple Roman */
+      else if ((platform == 1 && encoding == 0)	/* Apple Roman */
 	       || (platform == 2 && encoding == 0))
 	{			/* ISO ASCII */
 	  a->have_char_map_apple_roman = 1;
@@ -443,9 +525,13 @@ fontFetch (char **error, void *key)
   if (!found)
     {
       *error = "Unable to find a CharMap that I can handle";
+      /* 2.0.12: TBB: free these. Thanks to Frank Faubert. */
+      free (a->fontlist);
+      gdFree (a);
       return NULL;
     }
-
+  /* 2.0.5: we should actually return this */
+  a->face->charmap = found;
   return (void *) a;
 }
 
@@ -470,15 +556,17 @@ tweenColorTest (void *element, void *key)
 
   return (a->pixel == b->pixel
 	  && a->bgcolor == b->bgcolor
-	  && a->fgcolor == b->fgcolor
-	  && a->im == b->im);
+	  && a->fgcolor == b->fgcolor && a->im == b->im);
 }
 
 /*
  * Computes a color in im's color table that is part way between
  * the background and foreground colors proportional to the gray
  * pixel value in the range 0-NUMCOLORS. The fg and bg colors must already
- * be in the color table.
+ * be in the color table for palette images. For truecolor images the
+ * returned value simply has an alpha component and gdImageAlphaBlend
+ * does the work so that text can be alpha blended across a complex
+ * background (TBB; and for real in 2.0.2).
  */
 static void *
 tweenColorFetch (char **error, void *key)
@@ -497,7 +585,10 @@ tweenColorFetch (char **error, void *key)
   /* if fg is specified by a negative color idx, then don't antialias */
   if (fg < 0)
     {
-      a->tweencolor = -fg;
+      if ((pixel + pixel) >= NUMCOLORS)
+	a->tweencolor = -fg;
+      else
+	a->tweencolor = bg;
     }
   else
     {
@@ -509,18 +600,25 @@ tweenColorFetch (char **error, void *key)
 	     is incorporate our knowledge of the percentage of this
 	     pixel that is really "lit" by pushing the alpha value
 	     up toward transparency in edge regions. */
-	  a->tweencolor = gdTrueColorAlpha (
-					     gdTrueColorGetRed (fg),
-					     gdTrueColorGetGreen (fg),
-					     gdTrueColorGetBlue (fg),
-	       gdAlphaMax - (gdTrueColorGetAlpha (fg) * pixel / NUMCOLORS));
+	  a->tweencolor = gdTrueColorAlpha (gdTrueColorGetRed (fg),
+					    gdTrueColorGetGreen (fg),
+					    gdTrueColorGetBlue (fg),
+					    gdAlphaMax -
+					    (gdTrueColorGetAlpha (fg) *
+					     pixel / NUMCOLORS));
 	}
       else
 	{
 	  a->tweencolor = gdImageColorResolve (im,
-		   (pixel * im->red[fg] + npixel * im->red[bg]) / NUMCOLORS,
-	       (pixel * im->green[fg] + npixel * im->green[bg]) / NUMCOLORS,
-		(pixel * im->blue[fg] + npixel * im->blue[bg]) / NUMCOLORS);
+					       (pixel * im->red[fg] +
+						npixel * im->red[bg]) /
+					       NUMCOLORS,
+					       (pixel * im->green[fg] +
+						npixel * im->green[bg]) /
+					       NUMCOLORS,
+					       (pixel * im->blue[fg] +
+						npixel * im->blue[bg]) /
+					       NUMCOLORS);
 	}
     }
   return (void *) a;
@@ -534,30 +632,107 @@ tweenColorRelease (void *element)
 
 /* draw_bitmap - transfers glyph bitmap to GD image */
 static char *
-gdft_draw_bitmap (gdImage * im, int fg, FT_Bitmap bitmap, int pen_x, int pen_y)
+gdft_draw_bitmap (gdCache_head_t * tc_cache, gdImage * im, int fg,
+		  FT_Bitmap bitmap, int pen_x, int pen_y)
 {
-  unsigned char *pixel;
-  int *tpixel;
-  int x, y, row, col, pc;
+  unsigned char *pixel = NULL;
+  int *tpixel = NULL;
+  int x, y, row, col, pc, pcr;
 
   tweencolor_t *tc_elem;
   tweencolorkey_t tc_key;
 
-  /* initialize tweenColorCache on first call */
-  static gdCache_head_t *tc_cache;
-
-  if (!tc_cache)
-    {
-      tc_cache = gdCacheCreate (TWEENCOLORCACHESIZE,
-			tweenColorTest, tweenColorFetch, tweenColorRelease);
-    }
-
   /* copy to image, mapping colors */
   tc_key.fgcolor = fg;
   tc_key.im = im;
+  /* Truecolor version; does not require the cache */
+  if (im->trueColor)
+    {
+      for (row = 0; row < bitmap.rows; row++)
+	{
+	  pc = row * bitmap.pitch;
+	  pcr = pc;
+	  y = pen_y + row;
+	  /* clip if out of bounds */
+	  /* 2.0.16: clipping rectangle, not image bounds */
+	  if ((y > im->cy2) || (y < im->cy1))
+	    continue;
+	  for (col = 0; col < bitmap.width; col++, pc++)
+	    {
+	      int level;
+	      if (bitmap.pixel_mode == ft_pixel_mode_grays)
+		{
+		  /*
+		   * Scale to 128 levels of alpha for gd use.
+		   * alpha 0 is opacity, so be sure to invert at the end
+		   */
+		  level = (bitmap.buffer[pc] * gdAlphaMax /
+			   (bitmap.num_grays - 1));
+		}
+	      else if (bitmap.pixel_mode == ft_pixel_mode_mono)
+		{
+		  /* 2.0.5: mode_mono fix from Giuliano Pochini */
+		  level =
+		    ((bitmap.
+		      buffer[(col >> 3) +
+			     pcr]) & (1 << (~col & 0x07))) ?
+		    gdAlphaTransparent : gdAlphaOpaque;
+		}
+	      else
+		{
+		  return "Unsupported ft_pixel_mode";
+		}
+	      if ((fg >= 0) && (im->trueColor))
+		{
+		  /* Consider alpha in the foreground color itself to be an
+		     upper bound on how opaque things get, when truecolor is
+		     available. Without truecolor this results in far too many
+		     color indexes. */
+		  level =
+		    level * (gdAlphaMax -
+			     gdTrueColorGetAlpha (fg)) / gdAlphaMax;
+		}
+	      level = gdAlphaMax - level;
+	      x = pen_x + col;
+	      /* clip if out of bounds */
+	      /* 2.0.16: clip to clipping rectangle, Matt McNabb */
+	      if ((x > im->cx2) || (x < im->cx1))
+		continue;
+	      /* get pixel location in gd buffer */
+	      tpixel = &im->tpixels[y][x];
+	      if (fg < 0)
+		{
+		  if (level < (gdAlphaMax / 2))
+		    {
+		      *tpixel = -fg;
+		    }
+		}
+	      else
+		{
+		  if (im->alphaBlendingFlag)
+		    {
+		      *tpixel =
+			gdAlphaBlend (*tpixel,
+				      (level << 24) + (fg & 0xFFFFFF));
+		    }
+		  else
+		    {
+		      *tpixel = (level << 24) + (fg & 0xFFFFFF);
+		    }
+		}
+	    }
+	}
+      return (char *) NULL;
+    }
+  /* Non-truecolor case, restored to its more or less original form */
   for (row = 0; row < bitmap.rows; row++)
     {
+      int pcr;
       pc = row * bitmap.pitch;
+      pcr = pc;
+      if (bitmap.pixel_mode == ft_pixel_mode_mono)
+	pc *= 8;		/* pc is measured in bits for monochrome images */
+
       y = pen_y + row;
 
       /* clip if out of bounds */
@@ -581,61 +756,38 @@ gdft_draw_bitmap (gdImage * im, int fg, FT_Bitmap bitmap, int pen_x, int pen_y)
 	    {
 	      tc_key.pixel = ((bitmap.buffer[pc / 8]
 			       << (pc % 8)) & 128) ? NUMCOLORS : 0;
+	      /* 2.0.5: mode_mono fix from Giuliano Pochini */
+	      tc_key.pixel =
+		((bitmap.
+		  buffer[(col >> 3) +
+			 pcr]) & (1 << (~col & 0x07))) ? NUMCOLORS : 0;
 	    }
 	  else
 	    {
 	      return "Unsupported ft_pixel_mode";
 	    }
-
-	  if (tc_key.pixel > 0)
-	    {			/* if not background */
+	  if (tc_key.pixel > 0)	/* if not background */
+	    {
 	      x = pen_x + col;
 
 	      /* clip if out of bounds */
 	      if (x >= im->sx || x < 0)
 		continue;
 	      /* get pixel location in gd buffer */
-	      if (im->trueColor)
-		{
-		  tpixel = &im->tpixels[y][x];
-		}
-	      else
-		{
-		  pixel = &im->pixels[y][x];
-		}
+	      pixel = &im->pixels[y][x];
 	      if (tc_key.pixel == NUMCOLORS)
 		{
-		  /* use fg color directly */
-		  if (im->trueColor)
-		    {
-		      *tpixel = fg;
-		    }
-		  else
-		    {
-		      *pixel = fg;
-		    }
+		  /* use fg color directly. gd 2.0.2: watch out for
+		     negative indexes (thanks to David Marwood). */
+		  *pixel = (fg < 0) ? -fg : fg;
 		}
 	      else
 		{
 		  /* find antialised color */
-		  if (im->trueColor)
-		    {
-		      tc_key.bgcolor = *tpixel;
-		    }
-		  else
-		    {
-		      tc_key.bgcolor = *pixel;
-		    }
-		  tc_elem = (tweencolor_t *) gdCacheGet (
-							  tc_cache, &tc_key);
-		  if (im->trueColor)
-		    {
-		      *tpixel = tc_elem->tweencolor;
-		    }
-		  else
-		    {
-		      *pixel = tc_elem->tweencolor;
-		    }
+
+		  tc_key.bgcolor = *pixel;
+		  tc_elem = (tweencolor_t *) gdCacheGet (tc_cache, &tc_key);
+		  *pixel = tc_elem->tweencolor;
 		}
 	    }
 	}
@@ -643,14 +795,77 @@ gdft_draw_bitmap (gdImage * im, int fg, FT_Bitmap bitmap, int pen_x, int pen_y)
   return (char *) NULL;
 }
 
+static int
+gdroundupdown (FT_F26Dot6 v1, int updown)
+{
+  return (!updown)
+    ? (v1 < 0 ? ((v1 - 63) >> 6) : v1 >> 6)
+    : (v1 > 0 ? ((v1 + 63) >> 6) : v1 >> 6);
+}
+
 extern int any2eucjp (char *, char *, unsigned int);
 
+/* Persistent font cache until explicitly cleared */
+/* Fonts can be used across multiple images */
+
+/* 2.0.16: thread safety (the font cache is shared) */
+gdMutexDeclare (gdFontCacheMutex);
+static gdCache_head_t *fontCache;
+static FT_Library library;
+
+void
+gdFreeFontCache ()
+{
+  gdFontCacheShutdown ();
+}
+
+void
+gdFontCacheShutdown ()
+{
+  if (fontCache)
+    {
+      gdMutexShutdown (gdFontCacheMutex);
+      gdCacheDelete (fontCache);
+      FT_Done_FreeType (library);
+      /* 2.0.16: Gustavo Scotti: make sure we don't free this twice */
+      fontCache = 0;
+    }
+}
+
 /********************************************************************/
 /* gdImageStringFT -  render a utf8 string onto a gd image          */
 
 char *
 gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 		 double ptsize, double angle, int x, int y, char *string)
+{
+  return gdImageStringFTEx (im, brect, fg, fontlist,
+			    ptsize, angle, x, y, string, 0);
+}
+
+int
+gdFontCacheSetup (void)
+{
+  if (fontCache)
+    {
+      /* Already set up */
+      return 0;
+    }
+  gdMutexSetup (gdFontCacheMutex);
+  if (FT_Init_FreeType (&library))
+    {
+      gdMutexShutdown (gdFontCacheMutex);
+      return -1;
+    }
+  fontCache = gdCacheCreate (FONTCACHESIZE, fontTest, fontFetch, fontRelease);
+  return 0;
+}
+
+
+char *
+gdImageStringFTEx (gdImage * im, int *brect, int fg, char *fontlist,
+		   double ptsize, double angle, int x, int y, char *string,
+		   gdFTStringExtraPtr strex)
 {
   FT_BBox bbox, glyph_bbox;
   FT_Matrix matrix;
@@ -671,36 +886,75 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
   char *tmpstr = 0;
   int render = (im && (im->trueColor || (fg <= 255 && fg >= -255)));
   FT_BitmapGlyph bm;
+  /* 2.0.13: Bob Ostermann: don't force autohint, that's just for testing 
+     freetype and doesn't look as good */
+  int render_mode = FT_LOAD_DEFAULT;
+  int m, mfound;
+  /* Now tuneable thanks to Wez Furlong */
+  double linespace = LINESPACE;
+  /* 2.0.6: put this declaration with the other declarations! */
+  /*
+   *   make a new tweenColorCache on every call
+   *   because caching colormappings between calls
+   *   is not safe. If the im-pointer points to a
+   *   brand new image, the cache gives out bogus
+   *   colorindexes.          -- 27.06.2001 <krisku@arrak.fi>
+   */
+  gdCache_head_t *tc_cache;
+  /* Tuneable horizontal and vertical resolution in dots per inch */
+  int hdpi, vdpi;
+  if (strex)
+    {
+      if ((strex->flags & gdFTEX_LINESPACE) == gdFTEX_LINESPACE)
+	{
+	  linespace = strex->linespacing;
+	}
+    }
+  tc_cache = gdCacheCreate (TWEENCOLORCACHESIZE,
+			    tweenColorTest, tweenColorFetch,
+			    tweenColorRelease);
 
 /***** initialize font library and font cache on first call ******/
-  static gdCache_head_t *fontCache;
-  static FT_Library library;
-
   if (!fontCache)
     {
-      if (FT_Init_FreeType (&library))
+      if (gdFontCacheSetup () != 0)
 	{
+	  gdCacheDelete (tc_cache);
 	  return "Failure to initialize font library";
 	}
-      fontCache = gdCacheCreate (FONTCACHESIZE,
-				 fontTest, fontFetch, fontRelease);
     }
 /*****/
-
+  gdMutexLock (gdFontCacheMutex);
   /* get the font (via font cache) */
   fontkey.fontlist = fontlist;
   fontkey.library = &library;
   font = (font_t *) gdCacheGet (fontCache, &fontkey);
   if (!font)
     {
+      gdCacheDelete (tc_cache);
+      gdMutexUnlock (gdFontCacheMutex);
       return fontCache->error;
     }
   face = font->face;		/* shortcut */
   slot = face->glyph;		/* shortcut */
 
-  if (FT_Set_Char_Size (face, 0, (FT_F26Dot6) (ptsize * 64),
-			GD_RESOLUTION, GD_RESOLUTION))
+  /*
+   * Added hdpi and vdpi to support images at non-screen resolutions, i.e. 300 dpi TIFF,
+   *    or 100h x 50v dpi FAX format. 2.0.23.
+   * 2004/02/27 Mark Shackelford, mark.shackelford@acs-inc.com
+   */
+  hdpi = GD_RESOLUTION;
+  vdpi = GD_RESOLUTION;
+  if (strex && (strex->flags & gdFTEX_RESOLUTION))
     {
+      hdpi = strex->hdpi;
+      vdpi = strex->vdpi;
+    }
+
+  if (FT_Set_Char_Size (face, 0, (FT_F26Dot6) (ptsize * 64), hdpi, vdpi))
+    {
+      gdCacheDelete (tc_cache);
+      gdMutexUnlock (gdFontCacheMutex);
       return "Could not set character size";
     }
 
@@ -715,12 +969,61 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 
   use_kerning = FT_HAS_KERNING (face);
   previous = 0;
+  if (fg < 0)
+    {
+      render_mode |= FT_LOAD_MONOCHROME;
+    }
+  /* 2.0.12: allow explicit specification of the preferred map;
+     but we still fall back if it is not available. */
+  m = gdFTEX_Unicode;
+  if (strex && (strex->flags & gdFTEX_CHARMAP))
+    {
+      m = strex->charmap;
+    }
+  /* Try all three types of maps, but start with the specified one */
+  mfound = 0;
+  for (i = 0; (i < 3); i++)
+    {
+      switch (m)
+	{
+	case gdFTEX_Unicode:
+	  if (font->have_char_map_unicode)
+	    {
+	      mfound = 1;
+	    }
+	  break;
+	case gdFTEX_Shift_JIS:
+	  if (font->have_char_map_sjis)
+	    {
+	      mfound = 1;
+	    }
+	  break;
+	case gdFTEX_Big5:
+	  {
+	    /* This was the 'else' case, we can't really 'detect' it */
+	    mfound = 1;
+	  }
+	  break;
+	}
+      if (mfound)
+	{
+	  break;
+	}
+      m++;
+      m %= 3;
+    }
+  if (!mfound)
+    {
+      /* No character set found! */
+      gdMutexUnlock (gdFontCacheMutex);
+      return "No character set found";
+    }
 
 #ifndef JISX0208
   if (font->have_char_map_sjis)
     {
 #endif
-      if (tmpstr = (char *) gdMalloc (BUFSIZ))
+      if ((tmpstr = (char *) gdMalloc (BUFSIZ)))
 	{
 	  any2eucjp (tmpstr, string, BUFSIZ);
 	  next = tmpstr;
@@ -754,8 +1057,10 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
       /* newlines */
       if (ch == '\n')
 	{
-	  penf.y -= face->size->metrics.height * LINESPACE;
-	  penf.y = (penf.y - 32) & -64;		/* round to next pixel row */
+	  /* 2.0.13: reset penf.x. Christopher J. Grayce */
+	  penf.x = 0;
+	  penf.y -= face->size->metrics.height * linespace;
+	  penf.y = (penf.y - 32) & -64;	/* round to next pixel row */
 	  x1 = (penf.x * cos_a - penf.y * sin_a + 32) / 64;
 	  y1 = (penf.x * sin_a + penf.y * cos_a + 32) / 64;
 	  pen.x = pen.y = 0;
@@ -764,63 +1069,86 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 	  continue;
 	}
 
-      if (font->have_char_map_unicode)
+/* EAM DEBUG */
+#if (defined(FREETYPE_MAJOR) && (FREETYPE_MAJOR >=2 ) && (FREETYPE_MINOR >= 1))
+      if (font->face->charmap->encoding == FT_ENCODING_MS_SYMBOL)
 	{
-	  /* use UTF-8 mapping from ASCII */
+	  /* I do not know the significance of the constant 0xf000. */
+	  /* It was determined by inspection of the character codes */
+	  /* stored in Microsoft font symbol.ttf                    */
 	  len = gdTcl_UtfToUniChar (next, &ch);
+	  ch |= 0xf000;
 	  next += len;
 	}
-      else if (font->have_char_map_sjis)
-	{
-	  unsigned char c;
-	  int jiscode;
-
-	  c = *next;
-	  if (0xA1 <= c && c <= 0xFE)
-	    {
-	      next++;
-	      jiscode = 0x100 * (c & 0x7F) + ((*next) & 0x7F);
-
-	      ch = (jiscode >> 8) & 0xFF;
-	      jiscode &= 0xFF;
-
-	      if (ch & 1)
-		jiscode += 0x40 - 0x21;
-	      else
-		jiscode += 0x9E - 0x21;
-
-	      if (jiscode >= 0x7F)
-		jiscode++;
-	      ch = (ch - 0x21) / 2 + 0x81;
-	      if (ch >= 0xA0)
-		ch += 0x40;
-
-	      ch = (ch << 8) + jiscode;
-	    }
-	  else
-	    {
-	      ch = c & 0xFF;	/* don't extend sign */
-	    }
-	  next++;
-	}
       else
-	{
-	  /*
-	   * Big 5 mapping:
-	   * use "JIS-8 half-width katakana" coding from 8-bit characters. Ref:
-	   * ftp://ftp.ora.com/pub/examples/nutshell/ujip/doc/japan.inf-032092.sjs
-	   */
-	  ch = (*next) & 0xFF;	/* don't extend sign */
-	  next++;
-	  if (ch >= 161		/* first code of JIS-8 pair */
-	      && *next)
-	    {			/* don't advance past '\0' */
-	      /* TBB: Fix from Kwok Wah On: & 255 needed */
-	      ch = (ch * 256) + ((*next) & 255);
-	      next++;
-	    }
-	}
+#endif /* Freetype 2.1 or better */
+/* EAM DEBUG */
 
+	switch (m)
+	  {
+	  case gdFTEX_Unicode:
+	    if (font->have_char_map_unicode)
+	      {
+		/* use UTF-8 mapping from ASCII */
+		len = gdTcl_UtfToUniChar (next, &ch);
+		next += len;
+	      }
+	    break;
+	  case gdFTEX_Shift_JIS:
+	    if (font->have_char_map_sjis)
+	      {
+		unsigned char c;
+		int jiscode;
+		c = *next;
+		if (0xA1 <= c && c <= 0xFE)
+		  {
+		    next++;
+		    jiscode = 0x100 * (c & 0x7F) + ((*next) & 0x7F);
+
+		    ch = (jiscode >> 8) & 0xFF;
+		    jiscode &= 0xFF;
+
+		    if (ch & 1)
+		      jiscode += 0x40 - 0x21;
+		    else
+		      jiscode += 0x9E - 0x21;
+
+		    if (jiscode >= 0x7F)
+		      jiscode++;
+		    ch = (ch - 0x21) / 2 + 0x81;
+		    if (ch >= 0xA0)
+		      ch += 0x40;
+
+		    ch = (ch << 8) + jiscode;
+		  }
+		else
+		  {
+		    ch = c & 0xFF;	/* don't extend sign */
+		  }
+		next++;
+	      }
+	    break;
+	  case gdFTEX_Big5:
+	    {
+	      /*
+	       * Big 5 mapping:
+	       * use "JIS-8 half-width katakana" coding from 8-bit characters. Ref:
+	       * ftp://ftp.ora.com/pub/examples/nutshell/ujip/doc/japan.inf-032092.sjs
+	       */
+	      ch = (*next) & 0xFF;	/* don't extend sign */
+	      next++;
+	      if (ch >= 161	/* first code of JIS-8 pair */
+		  && *next)
+		{		/* don't advance past '\0' */
+		  /* TBB: Fix from Kwok Wah On: & 255 needed */
+		  ch = (ch * 256) + ((*next) & 255);
+		  next++;
+		}
+	    }
+	    break;
+	  }
+      /* set rotation transform */
+      FT_Set_Transform (face, &matrix, NULL);
       /* Convert character code to glyph index */
       glyph_index = FT_Get_Char_Index (face, ch);
 
@@ -830,53 +1158,67 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 	  FT_Get_Kerning (face, previous, glyph_index,
 			  ft_kerning_default, &delta);
 	  pen.x += delta.x;
+	  /* 2.0.12: indispensable for an accurate bounding box. */
+	  penf.x += delta.x;
 	}
 
       /* load glyph image into the slot (erase previous one) */
-      err = FT_Load_Glyph (face, glyph_index, FT_LOAD_DEFAULT);
+      err = FT_Load_Glyph (face, glyph_index, render_mode);
       if (err)
-	return "Problem loading glyph";
+	{
+	  gdCacheDelete (tc_cache);
+	  gdMutexUnlock (gdFontCacheMutex);
+	  return "Problem loading glyph";
+	}
 
       /* transform glyph image */
       FT_Get_Glyph (slot, &image);
       if (brect)
 	{			/* only if need brect */
 	  FT_Glyph_Get_CBox (image, ft_glyph_bbox_gridfit, &glyph_bbox);
-	  if (!i)
-	    {			/* if first character, init BB corner values */
-	      bbox.xMin = bbox.yMin = (1 << 30) - 1;
-	      bbox.xMax = bbox.yMax = -bbox.xMin;
-	    }
 	  glyph_bbox.xMin += penf.x;
 	  glyph_bbox.yMin += penf.y;
 	  glyph_bbox.xMax += penf.x;
 	  glyph_bbox.yMax += penf.y;
-	  if (bbox.xMin > glyph_bbox.xMin)
-	    bbox.xMin = glyph_bbox.xMin;
-	  if (bbox.yMin > glyph_bbox.yMin)
-	    bbox.yMin = glyph_bbox.yMin;
-	  if (bbox.xMax < glyph_bbox.xMax)
-	    bbox.xMax = glyph_bbox.xMax;
-	  if (bbox.yMax < glyph_bbox.yMax)
-	    bbox.yMax = glyph_bbox.yMax;
+	  if (ch == ' ')	/* special case for trailing space */
+	    glyph_bbox.xMax += slot->metrics.horiAdvance;
+	  if (!i)
+	    {			/* if first character, init BB corner values */
+	      bbox.xMin = glyph_bbox.xMin;
+	      bbox.yMin = glyph_bbox.yMin;
+	      bbox.xMax = glyph_bbox.xMax;
+	      bbox.yMax = glyph_bbox.yMax;
+	    }
+	  else
+	    {
+	      if (bbox.xMin > glyph_bbox.xMin)
+		bbox.xMin = glyph_bbox.xMin;
+	      if (bbox.yMin > glyph_bbox.yMin)
+		bbox.yMin = glyph_bbox.yMin;
+	      if (bbox.xMax < glyph_bbox.xMax)
+		bbox.xMax = glyph_bbox.xMax;
+	      if (bbox.yMax < glyph_bbox.yMax)
+		bbox.yMax = glyph_bbox.yMax;
+	    }
 	  i++;
 	}
 
-      /* transform glyph image */
-      FT_Glyph_Transform (image, &matrix, 0);
-
       if (render)
 	{
 	  if (image->format != ft_glyph_format_bitmap)
 	    {
 	      err = FT_Glyph_To_Bitmap (&image, ft_render_mode_normal, 0, 1);
 	      if (err)
-		return "Problem rendering glyph";
+		{
+		  gdCacheDelete (tc_cache);
+		  gdMutexUnlock (gdFontCacheMutex);
+		  return "Problem rendering glyph";
+		}
 	    }
 
 	  /* now, draw to our target surface */
 	  bm = (FT_BitmapGlyph) image;
-	  gdft_draw_bitmap (im, fg, bm->bitmap,
+	  gdft_draw_bitmap (tc_cache, im, fg, bm->bitmap,
 			    x + x1 + ((pen.x + 31) >> 6) + bm->left,
 			    y - y1 + ((pen.y + 31) >> 6) - bm->top);
 	}
@@ -922,15 +1264,9 @@ gdImageStringFT (gdImage * im, int *brect, int fg, char *fontlist,
 
   if (tmpstr)
     gdFree (tmpstr);
+  gdCacheDelete (tc_cache);
+  gdMutexUnlock (gdFontCacheMutex);
   return (char *) NULL;
 }
 
-int
-gdroundupdown (FT_F26Dot6 v1, int updown)
-{
-  return (!updown)
-    ? (v1 < 0 ? ((v1 - 63) >> 6) : v1 >> 6)
-    : (v1 > 0 ? ((v1 + 63) >> 6) : v1 >> 6);
-}
-
 #endif /* HAVE_LIBFREETYPE */
